"""
Defines the Card class which models a Magic: the Gathering card's behaviour.
Currently, the following actions are supported: to gain mana instanteneously as
well as according to a pattern every turn, to create gold/create gold/draw Cards
immediately as well as according to a pattern every turn. Additionally, a card
has a mana cost and can be a land or nonland. Finally, it may cause the removal of
a number of lands from the remaining deck.
"""

from typing import Optional
from mtg_mana_simulator.sequence import Sequence

class Card:
    """Simplified model of a Magic card"""

    untapped_land : "Card"
    tapped_land : "Card"
    cantrip : "Card"
    filler : "Card"

    def __init__(self, name: str = "", *,
            land: bool = False,
            cost: int = 0,
            mana_sequence: Optional[Sequence] = None,
            draw_sequence: Optional[Sequence] = None,
            gold_sequence: Optional[Sequence] = None,
            lands_removed: int = 0) -> None:
        self.name = name
        self.land = land
        self.cost = cost
        self.mana_sequence: Sequence = mana_sequence if mana_sequence is not None else Sequence.zero
        self.draw_sequence: Sequence = draw_sequence if draw_sequence is not None else Sequence.zero
        self.gold_sequence: Sequence = gold_sequence if gold_sequence is not None else Sequence.zero
        self.lands_removed = lands_removed

    def approximate_net_mana_sequence(self) -> Sequence:
        """ (assuming gold is spent immediately)"""
        return self.mana_sequence + self.gold_sequence + Sequence.once(-self.cost)

    def netgain(self) -> int:
        """The immediate return in mana upon playing this card"""
        mana = self.mana_sequence.finite_prefix(1)[0]
        gold = self.gold_sequence.finite_prefix(1)[0]
        return mana + gold - self.cost

    def is_ramp(self) -> bool:
        """Whether this card produces mana at some point"""
        return self.mana_sequence != Sequence.zero or self.gold_sequence != Sequence.zero

    def is_draw(self) -> bool:
        """Whether this card draws cards at some point"""
        return self.draw_sequence != Sequence.zero

    @staticmethod
    def untapped_rock(cost: int, mana: int) -> "Card":
        """Card with given cost and given mana gain"""
        return Card(cost=cost, mana_sequence=Sequence.repeat(mana))

    @staticmethod
    def tapped_rock(cost: int, mana: int) -> "Card":
        """Card with given cost and given mana gain starting next turn"""
        return Card(cost=cost, mana_sequence=Sequence.repeat(mana).prefixed_by([0]))

    @staticmethod
    def draw_spell(cost: int, cards: int) -> "Card":
        """Card with given cost that immediately draws given amount of cards"""
        return Card(cost=cost, draw_sequence=Sequence.once(cards))

Card.untapped_land = Card("Untapped land", land=True, mana_sequence=Sequence.one)
Card.tapped_land   = Card("Tapped land", land=True, mana_sequence=Sequence.one.prefixed_by([0]))
Card.cantrip       = Card.draw_spell(1, 1)
Card.filler        = Card("Filler", cost=20000)



"""
Defines the AI class which is the decision agent that decides whether to mulligan
or not, and which cards should be played given a certain 'context'.
"""

from typing import Callable, Iterator, List, Optional
from random import choice, shuffle
from mtg_mana_simulator.card import Card
from mtg_mana_simulator.context import Context
from mtg_mana_simulator.sequence import Sequence

Mulligan = Callable[[Context, int], Optional[List[int]]]
Choose   = Callable[[Context],      Optional[int]]

class AI:
    """
    Decision agent that implements mulligan and card playing algorithms
    """

    dud : "AI"
    naive : "AI"
    less_naive : "AI"
    never_mulligan  : Mulligan = lambda _,__: list(range(7))
    never_choose    : Choose   = lambda _: None
    randomly_choose : Choose   = lambda context: choice(context.playable_cards())

    def __init__(self, *,
            mulligan: Optional[Mulligan] = None,
            choose: Optional[Choose] = None) -> None:
        self.mulligan = mulligan if mulligan is not None else AI.never_mulligan
        self.choose   = choose   if choose   is not None else AI.never_choose

    def execute_mulligan(self, deck: List[Card]) -> Context:
        """
        Keep applying chosen mulligan strategy until a valid choice is made, or the hand is empty
        """
        for keepable_cards in reversed(range(1,8)):
            hand = deck[:7]
            remaining = deck[7:]

            card_indices = self.mulligan(Context(hand=hand, remaining=remaining), keepable_cards)
            if card_indices is not None and\
               len(card_indices)==keepable_cards and\
               len(card_indices)==len(set(card_indices)) and\
               all(0 <= index < 7 for index in card_indices):
                # If AI decided to keep, and the choice was valid, process the accepted cards
                accepted, rejected = [], []
                for index,card in enumerate(hand):
                    if index in card_indices:
                        accepted.append(card)
                    else:
                        rejected.append(card)
                return Context(hand=accepted, remaining=remaining+rejected)

            # Otherwise, shuffle and offer to mulligan again, with one less card
            shuffle(deck)
        return Context(hand=[], remaining=deck)

    def run(self, *,
            deck: List[Card],
            turns: int) -> List[int]:
        """
        Simulate playing given deck for some number of turns and return maximum mana per turn
        """

        context = self.execute_mulligan(deck[:])
        mana_generators: List[Iterator[int]] = []
        draw_generators: List[Iterator[int]] = [Sequence.one.generator()]
        gold_generators: List[Iterator[int]] = []

        mana_per_turn = turns*[0]
        for turn in range(turns):
            # Draw as much cards as you're supposed to this turn
            context.turn  = turn+1
            context.draw_cards(sum(generator.__next__() for generator in draw_generators))
            context.mana  = sum(generator.__next__() for generator in mana_generators)
            context.gold += sum(generator.__next__() for generator in gold_generators)
            context.land_for_turn = False

            # Determine maximum attainable mana before cards have been played, and set
            # the value of all following turns to that value. I.e. the player could
            # have decided to not do something this turn, but to postpone it to a later
            # turn. This is relevant to decrease variance when cards like Dark Ritual
            # appear in a deck.
            max_attainable_mana = max(mana_per_turn[turn], context.max_attainable_mana())
            mana_per_turn[turn:] = [max_attainable_mana]*(turns-turn)
            while True:
                playable_cards = context.playable_cards()
                if len(playable_cards) == 0:
                    # Pass the turn if there's no playable cards left
                    break

                chosen = self.choose(context)
                if chosen not in playable_cards:
                    # Pass the turn if the AI decides to stop playing
                    break

                # Play chosen card
                generators = context.play_card(context.hand.pop(chosen))
                mana_generators.append(generators['mana'])
                draw_generators.append(generators['draw'])
                gold_generators.append(generators['gold'])

                # Maximum attainable mana may have changed after drawing cards
                max_attainable_mana = max(mana_per_turn[turn], context.max_attainable_mana())
                mana_per_turn[turn:] = [max_attainable_mana]*(turns-turn)
        return mana_per_turn

    @staticmethod
    def minimum_land_mulligan(min_cards: int, min_lands: int) -> Mulligan:
        """
        Mulligan to at most some number of cards, whenever having less than some number of lands
        """
        def func(context, keepable_cards):
            if keepable_cards <= min_cards or\
               sum(card.land for card in context.hand) >= min_lands:
                # Keep the land cards of the dealt hand if the number of cards gets too low,
                # or the desired minimum number of lands is attained.
                lands, nonlands = [], []
                for index,card in enumerate(context.hand):
                    if card.land:
                        lands.append(index)
                    else:
                        nonlands.append(index)
                return (lands + nonlands)[:keepable_cards]
            # Mulligan
            return None
        return func

def improved_land_choice(context: Context) -> Optional[int]:
    """
    Play untapped land if needed, then ramp, then draw, then randomly choose a playable card
    """
    hand = context.hand
    mana = context.mana
    gold = context.gold
    playable_cards = context.playable_cards()

    # If there's a reason to play an untapped land, play it
    untapped_lands   = [k for k in playable_cards if hand[k].land and hand[k].netgain() != 0]
    almost_playables = [k for k in playable_cards if hand[k].cost == mana+gold+1]
    if len(almost_playables) > 0 and len(untapped_lands) > 0:
        return choice(untapped_lands)

    # Otherwise, if there are any ramp spells, play them
    ramp = [k for k in playable_cards if hand[k].is_ramp()]
    if len(ramp) > 0:
        return choice(ramp)

    # Otherwise, if there are any draw spells, play them
    draw = [k for k in playable_cards if hand[k].is_draw()]
    if len(draw) > 0:
        return choice(draw)

    # Otherwise, play randomly (tapped lands)
    return choice(playable_cards)

AI.dud        = AI()
AI.naive      = AI(choose=AI.randomly_choose)
AI.less_naive = AI(mulligan=AI.minimum_land_mulligan(5, 3), choose=improved_land_choice)



"""
Defines the Sequence class which provides a way to create generators of integers that
iterate a finite list followed by a repeating pattern, as well as logic to compose
them.
"""

from typing import List, Iterator
from math import ceil
from mtg_mana_simulator.helpers import divisors, lcm

class Sequence:
    """Infinite sequence of integers consisting of a finite prefix and a repeating pattern"""

    zero : "Sequence"
    one  : "Sequence"

    def __init__(self, prefix: List[int], pattern: List[int]) -> None:
        self.prefix = prefix
        self.pattern = pattern if len(pattern) > 0 else [0]
        self.normalize()

    def normalize(self) -> "Sequence":
        """Normalizes the sequence, such that comparison can be done elementwise"""
        # Reduce prefix size by 'rotating' the repeated pattern
        while len(self.prefix) > 0 and self.prefix[-1] == self.pattern[-1]:
            self.prefix = self.prefix[:-1]
            self.pattern = self.pattern[-1:]+self.pattern[:-1]

        # Reduce pattern size by identifying sublist that multiply
        for divisor in divisors(len(self.pattern)):
            times = len(self.pattern)//divisor
            if self.pattern == times*self.pattern[:divisor]:
                self.pattern = self.pattern[:divisor]
                break
        return self

    # Assumes the sequences are normalized
    def __eq__(self, other) -> bool:
        if not isinstance(other, Sequence):
            return NotImplemented
        return self.prefix == other.prefix and self.pattern == other.pattern

    def __add__(self, other: 'Sequence') -> 'Sequence':
        prefix_length = max([len(self.prefix), len(other.prefix)])
        pattern_length = lcm(len(self.pattern), len(other.pattern))
        prefix1 = self.finite_prefix(prefix_length + pattern_length)
        prefix2 = other.finite_prefix(prefix_length + pattern_length)
        summed: List[int] = list(map(sum, zip(prefix1, prefix2)))
        return Sequence(summed[:prefix_length], summed[prefix_length:]).normalize()

    def prefixed_by(self, additional_prefix: List[int]) -> 'Sequence':
        """Prefix an additional number to the sequence"""
        return Sequence(additional_prefix + self.prefix, self.pattern).normalize()

    def finite_prefix(self, length: int) -> List[int]:
        """Returns finite prefix of the sequence of the given length"""
        if length <= len(self.prefix):
            return self.prefix[:length]
        times = ceil((length - len(self.prefix))/len(self.pattern))
        return (self.prefix + times*self.pattern)[:length]

    def generator(self) -> Iterator[int]:
        """Returns a generator that iterates the sequence"""
        for element in self.prefix:
            yield element
        while True:
            for element in self.pattern:
                yield element

    @staticmethod
    def once(number: int) -> 'Sequence':
        """Returns a sequence of all zeroes prefixed by the given number"""
        return Sequence([number], []).normalize()

    @staticmethod
    def repeat(number: int) -> 'Sequence':
        """Returns a sequence infinitely repeating the given number"""
        return Sequence([], [number])

Sequence.zero = Sequence.repeat(0) # The all zeroes sequence: 0, 0, 0, ...
Sequence.one  = Sequence.repeat(1) # The all ones sequence:   1, 1, 1, ...



"""
Defines the Metric class which models functions that can be executed on traces.
"""

from typing import Callable, List, TypeVar
from statistics import mean, median, mode, variance
TYPE = TypeVar('TYPE')

class Metric:
    """Computations that can be executed on the amount of mana in a turn (across traces)"""

    identity : "Metric"
    mean : "Metric"
    median : "Metric"
    mode : "Metric"
    variance : "Metric"
    below_curve : "Metric"
    on_curve : "Metric"
    above_curve : "Metric"
    minimum : "Metric"
    maximum : "Metric"

    def __init__(self, name: str, func: Callable[[int, List[int]], TYPE]) -> None:
        self.name = name
        self.func = func

    def compute(self, traces: List[List[int]]) -> List[TYPE]:
        """Compute the metric on the given traces"""
        return [self.func(t+1, list(ms)) for t,ms in enumerate(zip(*traces))]

    @staticmethod
    def minimum_mana(min_mana: int) -> "Metric":
        """Probability of having at least the given amount of mana"""
        func = lambda t,ms: sum(m >= min_mana for m in ms)/len(ms)
        return Metric(f"≥{min_mana} mana", func)

    @staticmethod
    def percentile(fraction: float) -> "Metric":
        """Percentile score"""
        func = lambda t,ms: sorted(ms)[min(len(ms), max(0, round(len(ms)*fraction)-1))]
        return Metric(f"{fraction}th percentile", func)

Metric.identity    = Metric("Identity",     lambda t,ms: (t,ms))
Metric.mean        = Metric("Mean",         lambda t,ms: mean(ms))
Metric.median      = Metric("Median",       lambda t,ms: int(median(ms)))
Metric.mode        = Metric("Mode",         lambda t,ms: mode(ms))
Metric.variance    = Metric("Variance",     lambda t,ms: variance(ms))
Metric.below_curve = Metric("<'turn' mana", lambda t,ms: sum(m <  t for m in ms)/len(ms))
Metric.on_curve    = Metric("≥'turn' mana", lambda t,ms: sum(m >= t for m in ms)/len(ms))
Metric.above_curve = Metric(">'turn' mana", lambda t,ms: sum(m >  t for m in ms)/len(ms))
Metric.minimum     = Metric("Minimum",      lambda t,ms: min(ms))
Metric.maximum     = Metric("Maximum",      lambda t,ms: max(ms))



"""
Defines the Context class which is used to manage the state. AI decisions are based
solely on a given context, and playing a card modifies a context.
"""

from typing import Dict, Iterator, List, Optional
from random import sample
from mtg_mana_simulator.card import Card

class Context:
    """Snapshot of the state at a particular point in time"""

    def __init__(self, *,
            turn: int = 0,
            mana: int = 0,
            gold: int = 0,
            land_for_turn: bool = False,
            hand: Optional[List[Card]] = None,
            remaining: Optional[List[Card]] = None) -> None:
        self.turn = turn
        self.mana = mana
        self.gold = gold
        self.land_for_turn = land_for_turn
        self.hand: List[Card] = hand if hand is not None else []
        self.remaining: List[Card] = remaining if remaining is not None else []

        # Caching behaviours
        self.cached_playable_cards: Optional[List[int]] = None
        self.cached_max_attainable_mana: Optional[int] = None

    def lands_in_hand(self) -> List[Card]:
        """List of all land cards in hand"""
        return [card for card in self.hand if card.land]

    def nonlands_in_hand(self) -> List[Card]:
        """List of all nonland cards in hand"""
        return [card for card in self.hand if not card.land]

    def remove_lands(self, number: int) -> None:
        """Randomly removes a number of lands from the deck"""
        if number > 0:
            land_indices = sample([k for k,card in enumerate(self.remaining) if card.land], number)
            land_indices.sort(reverse=True)
            for land_index in land_indices:
                self.remaining.pop(land_index)

    def draw_cards(self, number: int) -> None:
        """Draw a number of cards"""
        if number > 0:
            self.cached_max_attainable_mana = None
            self.cached_playable_cards = None
            for _ in range(number):
                self.hand.append(self.remaining.pop())

    def play_card(self, card: Card) -> Dict[str, Iterator[int]]:
        """Update the context by playing a given card"""
        self.cached_playable_cards = None
        generators = {
            'mana': card.mana_sequence.generator(),
            'draw': card.draw_sequence.generator(),
            'gold': card.gold_sequence.generator()
        }
        self.remove_lands(card.lands_removed)
        self.land_for_turn = self.land_for_turn or card.land
        self.gold += generators['gold'].__next__() - min([self.gold, max([0, card.cost-self.mana])])
        self.mana += generators['mana'].__next__() - min([self.mana, card.cost])
        self.draw_cards(generators['draw'].__next__())
        return generators

    def playable_cards(self) -> List[int]:
        """Returns a list of indices of the cards that can currently be played"""
        if self.cached_playable_cards is None:
            mana = self.mana+self.gold
            condition = lambda card: card.cost <= mana and not (card.land and self.land_for_turn)
            self.cached_playable_cards = [k for k,card in enumerate(self.hand) if condition(card)]
        return self.cached_playable_cards

    def max_attainable_mana(self) -> int:
        """Maximum attainable mana based on public information (lands and net gain)"""
        if self.cached_max_attainable_mana is None:
            max_attainable_mana = self.mana + self.gold
            if not self.land_for_turn:
                max_attainable_mana += max([0]+[card.netgain() for card in self.lands_in_hand()])
            for card in sorted(self.nonlands_in_hand(), key=lambda card: card.cost):
                max_attainable_mana += max(0, card.netgain())
                # Cards are traversed in cost order, so we can stop early if we can't pay the cost
                if card.cost > max_attainable_mana:
                    break
            self.cached_max_attainable_mana = max_attainable_mana
        return self.cached_max_attainable_mana



"""
mtg_mana_simulator allows you to quickly simulate Magic: the Gathering plays from a single
player's perspective to estimate certain measures related to available mana throughout the game.
"""

from mtg_mana_simulator.ai import AI
from mtg_mana_simulator.card import Card
from mtg_mana_simulator.context import Context
from mtg_mana_simulator.experiment import Experiment
from mtg_mana_simulator.metric import Metric
from mtg_mana_simulator.sequence import Sequence
from mtg_mana_simulator.version import __version__



"""
Installation script for this package
"""

from setuptools import setup, find_packages

setup(
    name='mtg-mana-simulator',
    version='0.2',
    url='https://github.com/TiesWestendorp/MTG-Mana-Simulator',
    author='Ties Westendorp',
    packages=find_packages(include=['mtg_mana_simulator', 'mtg_mana_simulator.*']),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.0',
)



"""
Helper methods
"""

from typing import List
from math import gcd, sqrt

def divisors(number: int) -> List[int]:
    """Returns a list of all divisors of the input"""
    divs = [1]
    for candidate in range(2,int(sqrt(number))+1):
        if number%candidate == 0:
            divs.extend([candidate,number//candidate])
    divs.extend([number])
    return list(set(divs))

def lcm(number1: int, number2: int) -> int:
    """Compute the least common multiple of two numbers"""
    return number1*number2//gcd(number1, number2)



"""
Defines the Experiment class which is the main class of this module.
"""

from typing import Any, Dict, List, Optional
from random import shuffle
from mtg_mana_simulator.ai import AI
from mtg_mana_simulator.card import Card
from mtg_mana_simulator.metric import Metric

class Experiment:
    """Wrapper class for simulating multiple runs"""

    def __init__(self,
            deck: List[Card],
            ai: AI,
            turns: int,
            repeats: int,
            *, options: Optional[Dict[str, Any]]=None) -> None:
        self.deck    = deck
        self.ai      = ai
        self.turns   = turns
        self.repeats = repeats
        self.options = options if options is not None else {}
        self.traces : List[List[int]] = self.repeats*[[0]]
        self.run()

    def run(self) -> None:
        """Run the experiment"""
        variance_reduction = None
        if isinstance(self.options, dict):
            if 'variance_reduction' in self.options:
                variance_reduction = self.options['variance_reduction']

        for iteration in range(self.repeats):
            if variance_reduction == 'antithetic-variates':
                # https://en.wikipedia.org/wiki/Antithetic_variates
                if iteration%2 == 0:
                    shuffle(self.deck)
                else:
                    self.deck.reverse()
            else:
                shuffle(self.deck)

            self.traces[iteration] = self.ai.run(deck=self.deck, turns=self.turns)

    def evaluate(self, metrics: List[Metric]) -> Dict[str, List[Any]]:
        """Evaluate given metrics on the generated traces"""
        return dict(map(lambda metric: (metric.name, metric.compute(self.traces)), metrics))
